import {Component, EventEmitter, Input, Output, ViewChild} from '@angular/core'
import {
  Mix,
  MixMaterial,
  MixMaterialDto,
  mixMaterialsAsMixMaterialsDto,
  Recipe,
  sortMixMaterialsDto
} from '../../../shared/model/recipe.model'
import {ErrorHandlingComponent} from '../../../shared/components/subscribing.component'
import {MixService} from '../../services/mix.service'
import {RecipeService} from '../../services/recipe.service'
import {Material, materialComparator} from '../../../shared/model/material.model'
import {MaterialService} from '../../../material/service/material.service'
import {MaterialTypeService} from '../../../material-type/service/material-type.service'
import {FormBuilder, FormControl, Validators} from '@angular/forms'
import {UNIT_MILLILITER} from '../../../shared/units'
import {MatTable} from '@angular/material/table'
import {ActivatedRoute, Router} from '@angular/router'
import {ConfirmBoxComponent} from '../../../shared/components/confirm-box/confirm-box.component'
import {AccountService} from '../../../accounts/services/account.service'
import {ErrorService} from '../../../shared/service/error.service'
import {map} from 'rxjs/operators';
import {CreInputEntry} from '../../../shared/components/inputs/inputs';
import {CreForm, ICreForm} from '../../../shared/components/forms/forms';

@Component({
  selector: 'cre-mix-editor',
  templateUrl: './mix-editor.component.html',
  styleUrls: ['./mix-editor.component.sass']
})
export class MixEditorComponent extends ErrorHandlingComponent {
  @ViewChild(MatTable, {static: true}) mixTable: MatTable<MixMaterial>
  @ViewChild('deleteConfirmBox') deleteConfirmBox: ConfirmBoxComponent
  @ViewChild(CreForm) creForm: ICreForm

  @Input() mixId: number | null
  @Input() recipe: Recipe
  @Input() mix: Mix | null
  @Input() materials: Material[] = []

  @Output() save = new EventEmitter<any>()

  editionMode = false
  units = UNIT_MILLILITER
  hoveredMixMaterial: MixMaterial | null
  tableColumns = ['position', 'buttonsPosition', 'material', 'quantity', 'units', 'buttonRemove']

  deleting = false
  errorHandlers = [{
    filter: error => error.type === 'notfound-mix-id',
    consumer: _ => this.urlUtils.navigateTo('/color/list')
  }, {
    filter: error => error.type === 'exists-material-name',
    messageProducer: error => `Un produit avec le nom '${error.name}' existe déjà`
  }, {
    filter: error => error.type === 'cannotdelete-mix',
    messageProducer: _ => 'Ce mélange est utilisé par un ou plusieurs autres mélanges'
  }, {
    filter: error => error.type === 'invalid-mixmaterial-first',
    messageProducer: _ => 'La quantité du premier ingrédient du mélange ne peut pas être exprimée en pourcentage'
  }]

  controls: any
  materialTypeEntries$ = this.materialTypeService.all.pipe(
    map(materialTypes => materialTypes.map(materialType => new CreInputEntry(materialType.id, materialType.name))),
  )

  private _mixMaterials: MixMaterialDto[] = []

  constructor(
    private mixService: MixService,
    private recipeService: RecipeService,
    private materialService: MaterialService,
    private materialTypeService: MaterialTypeService,
    private accountService: AccountService,
    private formBuilder: FormBuilder,
    errorService: ErrorService,
    router: Router,
    activatedRoute: ActivatedRoute
  ) {
    super(errorService, activatedRoute, router)
  }

  ngOnInit() {
    super.ngOnInit()

    this.mixId = this.urlUtils.parseIntUrlParam('id')
    if (this.mixId) {
      this.editionMode = true
    }

    if (this.editionMode) {
      this.mix = this.recipe.mixes.find(mix => mix.id === this.mixId)
      this.mixMaterials = mixMaterialsAsMixMaterialsDto(this.mix)
    } else {
      this.addBlankMixMaterial()
    }

    this.createControls()
  }

  private createControls() {
    this.controls = {
      name: new FormControl(this.mix?.mixType.name, Validators.required),
      materialType: new FormControl(this.mix?.mixType.material.materialType.id, Validators.required)
    }
  }

  addRow() {
    this.addBlankMixMaterial()
    this.mixTable.renderRows()
  }

  removeRow(position: number) {
    this.mixMaterials.splice(position, 1)

    // Decreases the position of each mix material above the removed one
    for (let i = position; i < this.mixMaterials.length; i++) {
      this.mixMaterials[i].position -= 1
    }

    this.mixTable.renderRows()
  }

  increasePosition(mixMaterial: MixMaterialDto, table: MatTable<any>) {
    this.updateMixMaterialPosition(mixMaterial, mixMaterial.position + 1)
    this.sort(table)
  }

  decreasePosition(mixMaterial: MixMaterialDto, table: MatTable<any>) {
    this.updateMixMaterialPosition(mixMaterial, mixMaterial.position - 1)
    this.sort(table)
  }

  sort(table: MatTable<any>) {
    this.mixMaterials = sortMixMaterialsDto(this.mixMaterials)
    table.renderRows()
  }

  setMixMaterialMaterial(mixMaterial: MixMaterialDto, materialId: number) {
    mixMaterial.isPercents = this.materials.find(m => m.id === materialId).materialType.usePercentages
    mixMaterial.materialId = materialId
  }

  submit() {
    this.save.emit({
      name: this.controls.name.value,
      recipeId: this.recipe.id,
      materialTypeId: this.controls.materialType.value,
      mixMaterials: this.mixMaterials,
      units: this.units
    })
  }

  delete() {
    this.deleting = true
    this.subscribeAndNavigate(this.mixService.delete(this.mixId), `/color/edit/${this.recipe.id}`)
  }

  getAvailableMaterials(mixMaterial: MixMaterialDto): CreInputEntry[] {
    // return this.materialService.all.pipe(
    //   map(materials => materials.filter(material => {
    //     return mixMaterial.materialId === material.id || this.mixMaterials.filter(mm => mm.materialId === material.id).length === 0
    //   })),
    //   map(materials => this.sortedMaterials(materials)),
    //   map(materials => materials.map(material => new CreInputEntry(material.id, material.name)))
    // )
    return this.materials
      .filter(m => mixMaterial.materialId === m.id || this.mixMaterials.filter(mm => mm.materialId === m.id).length === 0)
      .sort(materialComparator)
      .map(material => new CreInputEntry(material.id, material.name))
    // return this.materials.map(material => new CreInputEntry(material.id, material.name))
    // return this.materials
    //   .filter(m => mixMaterial.materialId === m.id || this.mixMaterials.filter(mm => mm.materialId === m.id).length === 0)
    //   .sort(materialComparator)
    //   .map(material => new CreInputEntry(material.id, material.name))
  }

  materialDisplayName(material: Material): string {
    if (material.materialType.prefix) {
      return `[${material.materialType.prefix}] ${material.name}`
    }
    return material.name
  }

  get mixMaterials(): MixMaterialDto[] {
    return this._mixMaterials
  }

  set mixMaterials(mixMaterials: MixMaterialDto[]) {
    this._mixMaterials = mixMaterials
    this.mixTable.renderRows()
  }

  private addBlankMixMaterial() {
    const mixMaterial = new MixMaterialDto(null, 0, false, this.mixMaterials.length + 1)
    this.mixMaterials = [...this.mixMaterials, mixMaterial]
  }

  private updateMixMaterialPosition(mixMaterial: MixMaterialDto, updatedPosition: number) {
    if (!this.mixMaterialAtPosition(updatedPosition)) {
      mixMaterial.position = updatedPosition
    } else {
      const conflictingStep = this.mixMaterialAtPosition(updatedPosition)
      conflictingStep.position = mixMaterial.position
      mixMaterial.position = updatedPosition
    }
  }

  private mixMaterialAtPosition(position: number): MixMaterialDto {
    return this.mixMaterials.find(m => m.position === position)
  }
}
